/* RootServ routines

   Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/services.h"

AutoKill *akilllist[1024];
Trigger *triggerlist = NULL;
Exception *exceptionlist = NULL;

static void do_help (const char *source);
static void do_global (const char *source);
static void do_jupe (const char *source);
static void do_clones (const char *source);
static void do_akill (const char *source);
static void do_akill_add (const char *source);
static void do_akill_del (const char *source);
static void do_akill_list (const char *source);
static void do_trigger (const char *source);
static void do_trigger_add (const char *source);
static void do_trigger_del (const char *source);
static void do_trigger_list (const char *source);
static void do_exception (const char *source);
static void do_exception_add (const char *source);
static void do_exception_del (const char *source);
static void do_exception_list (const char *source);
static void do_csop (const char *source);
static void do_csop_add (const char *source);
static void do_csop_del (const char *source);
static void do_csop_list (const char *source);
static void do_auth (const char *source);
static void do_deauth (const char *source);
#ifdef RAWINJECT
static void do_raw (const char *source);
static void do_inject (const char *source);
#endif
static void do_debug (const char *source);
static void do_shutdown (const char *source);
static void do_restart (const char *source);
static void do_settings (const char *source);
static void do_update (const char *source);
static void do_resetnews (const char *source);
static void do_search (const char *source);
static void do_stats (const char *source);

/* Temporary debugging functions. */
static void do_userlist (const char *source);
static void do_chanlist (const char *source);

/* Add a new trigger to the trigger list */
Trigger *new_trigger (const char *host, int limit)
{
    Trigger *trigger;
    char *username, *hostname;

    username = sstrdup (host);
    hostname = strchr (username, '@');

    if (!hostname)
	return NULL;

    *hostname++ = 0;

    if ((trigger = findtrig (username, hostname)))
    {
	free (username);
	return trigger;
    }

    trigger = scalloc (sizeof (Trigger), 1);

    strscpy (trigger->host, host, HOSTLEN);
    trigger->limit = limit;

    trigger->next = triggerlist;

    if (trigger->next)
        trigger->next->prev = trigger;

    triggerlist = trigger;

    free (username);

    return trigger;
}

/* Search for a trigger */
Trigger *findtrig (const char *user, const char *host)
{
    Trigger *trigger;
    char *mask;

    mask = smalloc (strlen (user) + strlen (host) + 2);
    sprintf (mask, "%s@%s", user, host);

    for (trigger = triggerlist; trigger; trigger = trigger->next)
	if (match_wild_nocase (trigger->host, mask))
	{
	    free (mask);
	    return trigger;
	}

    free (mask);
    return NULL;
}

/* Add a new exception to the exception list */
Exception *new_exception (const char *host)
{
    Exception *exception;
    char *username, *hostname;

    username = sstrdup (host);
    hostname = strchr (username, '@');

    if (!hostname)
	return NULL;

    *hostname++ = 0;

    if ((exception = findexcept (username, hostname)))
    {
	free (username);
	return exception;
    }

    exception = scalloc (sizeof (Exception), 1);

    strscpy (exception->host, host, HOSTLEN);

    exception->next = exceptionlist;

    if (exception->next)
	exception->next->prev = exception;

    exceptionlist = exception;

    free (username);

    return exception;
}

/* Search for an exception */
Exception *findexcept (const char *user, const char *host)
{
    Exception *exception;
    char *mask;

    mask = smalloc (strlen (user) + strlen (host) + 2);
    sprintf (mask, "%s@%s", user, host);

    for (exception = exceptionlist; exception; exception = exception->next)
	if (match_wild_nocase (exception->host, mask))
	{
	    free (mask);
	    return exception;
	}

    free (mask);

    return NULL;
}   

/* Global Noticer routine. This is simply to reply to CTCPs. */
void globalnoticer (const char *source, char *buf)
{
    char *cmd, *s;  

    cmd = strtok (buf, " ");

    if (!cmd)
	return;

    if (!stricmp (cmd, "\1PING"))
    {
	if (!(s = strtok (NULL, "")))
	    s = "0";
	strip (s);
	ctcpreply (s_GlobalNoticer, source, "PING %s", s);
	return;
    }
    else if (!stricmp (cmd, "\1VERSION\1"))
    {
	ctcpreply (s_GlobalNoticer, source, "VERSION Cygnus IRC Services v%s",
	    version);
	return;
    }
}

/* Main RootServ routine. */
void rootserv (const char *source, char *buf)  
{
    char *cmd, *s;
    char orig[BUFSIZE];

    /* Make a copy of the original message. */ 
    strscpy (orig, buf, sizeof (orig));   

    cmd = strtok (buf, " ");

    if (!cmd)
	return;
    
    if (!stricmp (cmd, "\1PING"))
    {
	if (!(s = strtok (NULL, "")))
	    s = "0";
	strip (s);
	ctcpreply (s_RootServ, source, "PING %s", s);
	return;
    }
    else if (!stricmp (cmd, "\1VERSION\1"))
    {
	ctcpreply (s_RootServ, source, "VERSION Cygnus IRC Services v%s",
	    version);
	return;
    }

    /* A CTCP we don't recognize.. Ignore it. */
    else if (*cmd == '\1')
	return;

    else if (!is_oper (source))
    {
	notice (s_RootServ, source, RS_ACCESS_DENIED);
	return;
    }

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP",		H_IRCOP,	do_help},
	    {"GLOBAL",		H_IRCOP,	do_global},
	    {"JUPE",		H_IRCOP,	do_jupe},
	    {"CLONES",		H_IRCOP,	do_clones},
	    {"AKILL",		H_IRCOP,	do_akill},
	    {"TRIGGER",		H_IRCOP,	do_trigger},
	    {"EXCEPTION",	H_IRCOP,	do_exception},
	    {"STATS",		H_IRCOP,	do_stats},
	    {"AUTH",		H_IRCOP,	do_auth},
	    {"SEARCH",		H_CSOP,		do_search},
	    {"DEAUTH",		H_SRA,		do_deauth},
	    {"CSOP",		H_SRA,		do_csop},
	    {"SHUTDOWN",	H_SRA,		do_shutdown},
	    {"RESTART",		H_SRA,		do_restart},
#ifdef RAWINJECT
	    {"RAW",		H_SRA,		do_raw},
	    {"INJECT",		H_SRA,		do_inject},
#endif
	    {"DEBUG",		H_SRA,		do_debug},
	    {"UPDATE",		H_SRA,		do_update},
	    {"REHASH",		H_SRA,		do_rehash},
	    {"SETTINGS",	H_SRA,		do_settings},
	    {"RESETNEWS",	H_SRA,		do_resetnews},
	    {"USERLIST",	H_SRA,		do_userlist},
	    {"CHANLIST",	H_SRA,		do_chanlist},
	    {NULL}
	};
            
	if ((command = get_hash (s_RootServ, source, strupper (cmd),
				 hash_table)))
	{
	    /* Check if they want a disabled command. */
	    if ((command->process == do_global &&
		globalnoticer_on == FALSE) ||
		(command->process == do_clones &&
		!maxclones) ||
		(command->process == do_trigger &&
		!maxclones) ||
		(command->process == do_exception &&
		!maxclones) ||
		(command->process == do_resetnews &&
		memoserv_on == FALSE))
	    {
		notice (s_RootServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd),
		    haveserv_on == TRUE ? "" : "MSG ", s_RootServ,
		    haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		    haveserv_on == TRUE ? "" : securitysetting == 1 ?
		    me.name : "");
		return;
	    }

	    (*command->process) (source);

	    if (!(command->process == do_auth)) /* No pw snooping */
	    {
		log ("RS: <%s> %s", source, orig);
		snoop (s_RootServ, "RS: <%s> %s", source, orig);
	    }
	    else /* But we do want to snoop something. */
	    {
		log ("RS: <%s> AUTH ...", source);
		snoop (s_RootServ, "RS: <%s> AUTH ...", source);
	    }
	}
    }
}

/* Return a help message. */
static void do_help (const char *source) 
{
    char *cmd = strtok (NULL, " ");
            
    if (!cmd)
	rootserv_help_index (source);
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"GLOBAL",		H_IRCOP,	rootserv_help_global},
	    {"JUPE",		H_IRCOP,	rootserv_help_jupe},
	    {"CLONES",		H_IRCOP,	rootserv_help_clones},
	    {"AKILL",		H_IRCOP,	rootserv_help_akill},
	    {"TRIGGER",		H_IRCOP,	rootserv_help_trigger},
	    {"EXCEPTION",	H_IRCOP,	rootserv_help_exception},
	    {"STATS",		H_IRCOP,	rootserv_help_stats},
	    {"AUTH",		H_IRCOP,	rootserv_help_auth},
	    {"SEARCH",		H_CSOP,		rootserv_help_search},
	    {"DEAUTH",		H_SRA,		rootserv_help_deauth},
	    {"SHUTDOWN",	H_SRA,		rootserv_help_shutdown},
	    {"RESTART",		H_SRA,		rootserv_help_restart},
#ifdef RAWINJECT
	    {"RAW",		H_SRA,		rootserv_help_raw},
	    {"INJECT",		H_SRA,		rootserv_help_inject},
#endif
	    {"CSOP",		H_SRA,		rootserv_help_csop},
	    {"DEBUG",		H_SRA,		rootserv_help_debug},
	    {"UPDATE",		H_SRA,		rootserv_help_update},
	    {"REHASH",		H_SRA,		rootserv_help_rehash},
	    {"SETTINGS",	H_SRA,		rootserv_help_settings},
	    {"RESETNEWS",	H_SRA,		rootserv_help_resetnews},
	    {NULL}
	};

	if ((command = get_help_hash (s_RootServ, source, strupper (cmd),
				      hash_table)))
	{
	    /* Check if they want a disabled command. */
	    if (((command->process == rootserv_help_global) && globalnoticer_on == FALSE) ||
		((command->process == rootserv_help_resetnews) && memoserv_on == FALSE) ||
		((command->process == rootserv_help_clones) && !maxclones == FALSE) ||
		((command->process == rootserv_help_trigger) && !maxclones == FALSE) ||
		((command->process == rootserv_help_exception) && !maxclones == FALSE))
	    {
		notice (s_RootServ, source, ERR_NO_HELP, strupper (cmd));
		return;
	    }

	    (*command->process) (source);
	}
    }
}

/* Jupiter a server or nickname. */
static void do_jupe (const char *source)
{
    char *param = strtok (NULL, " ");
    char *reason = strtok (NULL, "");

    if (!param)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "JUPE Server|Nick [Reason]");
	errmoreinfo (s_RootServ, source, "JUPE");
	return;
    }

    /* Sanity checks .. We could do these all in one shot, but the output
       (and resultant warnings) are more informative this way.
     */
    if (!stricmp (me.name, param))
    {
	notice (s_RootServ, source, ERR_NOT_SERVICES);
	globops (s_RootServ, RS_JUPE_MY_SERVER, source);
	return;
    }

    if (!stricmp (me.ruplink, param))
    {
	notice (s_RootServ, source, ERR_NOT_SERVICES);
	globops (s_RootServ, RS_JUPE_MY_UPLINK, source);
	return;
    }

    if (is_one_of_mine (param))
    {
	notice (s_RootServ, source, ERR_NOT_SERVICES);
	globops (s_RootServ, RS_JUPE_MY_NICK, source, param);
	return;
    }

    /* Check for a *. Rather than actually making a server called bob.* or
       soemthing, we'll bail out here. We could try matching against existing
       server names, but this is fairly 'dangerous', so we'll require a full
       server name.
     */
    if (strchr (param, '*'))
    {
	notice (s_RootServ, source, ERR_BAD_SERVER);
	return;
    }

    /* Is it a server? */
    if (strchr (param, '.'))
    {
	int i;

	/* Does that server already exist? */
	for (i = 0; i < servcnt; i++)
	{
	    if (!stricmp (servlist[i].server, param))
	    {
		/* Yep. Now, we squit it from the network, and free the
		   memory for it. */
		send_cmd (me.name, "%s %s :Juped by %s: %s", me.token ? "-" : 
		    "SQUIT", param, source, reason ? reason :
		    "No reason given");
		--servcnt;
		if (servlist[i].server)
		    free (servlist[i].server);
		if (i < servcnt)
		    bcopy (servlist + i + 1, servlist + i,
			sizeof (*servlist) * (servcnt - i));
	    }
	}

	/* Now the server is either gone, or was never there, so we bring
	   on the jupe.
	 */
#ifdef IRCUP9
	send_cmd (NULL, "SERVER %s 2 %lu %lu P09 :%s - %s", param,
	    time (NULL), time (NULL), reason ? reason : "Jupitered Server",
	    source);
#else
	send_cmd (NULL, "%s %s 2 :%s - %s", me.token ? "'" : "SERVER",
	    param, reason ? reason : "Jupitered Server", source);
#endif /* IRCUP9 */
	globops (s_RootServ, RS_JUPED, param, source);
	notice (s_RootServ, source, RS_JUPE_SUCCESSFUL, param);
    }

    /* Doesn't appear to be a server, treat it as a nickname. */
    else
    {
	User *u;
	char buf[256];

	/* Is the nick in use? */
	if ((u = finduser (param)))
	{
	    /* Yep. Now we kill the client before bringing on the jupe. */
	    sprintf (buf, "Jupitered by %s: %s", source,
		reason ? reason : "No reason given");
	    kill_user (s_RootServ, u->nick, buf);
	}

	/* Now the nick is gone, or was never there, so we bring on the
	   jupe.
	 */
	sprintf (buf, "Jupitered Nick - %s", source);
	intro_user (param, "jupe", rs.host, buf, "+");
	globops (s_RootServ, RS_JUPED, param, source);
	notice (s_RootServ, source, RS_JUPE_SUCCESSFUL, param);
    }
}

/* Send a global notice */
static void do_global (const char *source)
{
    char *text = strtok (NULL, ""), buf[BUFSIZE];

    if (!text)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "GLOBAL Message");
	errmoreinfo (s_RootServ, source, "GLOBAL");
	return;
    }

    if (strlen (globalmsg))
    {
	*buf = 0;

	strcat (buf, globalmsg);
	strcat (buf, " ");
	strcat (buf, text);

	noticeall (buf);

	globops (s_RootServ, RS_GLOBAL_NOTICE, source);

	return;
    }

    noticeall (text);
    globops (s_RootServ, RS_GLOBAL_NOTICE, source);
}

/* Reset NF_NEWS flags for all nicks. */
static void do_resetnews (const char *source)
{
    NickInfo *ni;
    int i;

    globops (s_RootServ, RS_RESETTING_NEWS, source);

    /* Run through the registered nicks, resetting the flag for each. */
    for (i = 0; i < 1024; i++)
	for (ni = nicklist[i]; ni; ni = ni->next)
	    if (!(ni->flags & NF_LINKED))
		ni->flags |= NF_NEWS;

    /* Now set the time of when we reset it. */
    news_time = time (NULL);

    notice (s_RootServ, source, RS_NEWS_RESET);
}

/* Handle an AKILL command */
static void do_akill (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_RootServ, source, RPL_SYNTAX,
	    "AKILL ADD|DEL|LIST [Mask|Number|Option]");
	errmoreinfo (s_RootServ, source, "AKILL");
    }
    else
    {
	Hash_SUB *command, hash_table[] =
	{
	    {"ADD",		H_IRCOP,	do_akill_add},
	    {"DEL",		H_IRCOP,	do_akill_del},
	    {"LIST",		H_IRCOP,	do_akill_list},
	    {NULL}
	};

	if ((command = get_sub_hash (s_RootServ, source, strupper (cmd),
				 hash_table)))
	    (*command->process) (source);
	else
	{
	    notice (s_RootServ, source, RPL_SYNTAX,
		"AKILL ADD|DEL|LIST [Mask|Number|Option]");
	    errmoreinfo (s_RootServ, source, "AKILL");
	}
    }
}

/* Add an AKill to the AKill list */
void do_akill_add (const char *source)
{
    AutoKill *akill, **list;
    User *u;
    char *param, mask[BUFSIZE], reason[BUFSIZE], *tmp, orig[BUFSIZE], *username;
    char *hostname;
    int i, j = 0, nonwild = 0, realname = 0, perm = 0, expires;

    param = strtok (NULL, " ");
    tmp = strtok (NULL, "");

    if (!param)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "AKILL ADD Mask [Expiry] Reason");
	errmoreinfo (s_RootServ, source, "AKILL");
	return;
    }

    if (!tmp)
	memcpy (reason, def_akill_reason, sizeof (reason));
    else
	memcpy (reason, tmp, sizeof (reason));

    strscpy (orig, reason, sizeof (orig));

    /* If we don't find an expire time below, we'll use this default */
    expires = def_akill_time;

    tmp = strtok (reason, " ");

    if (!stricmp (tmp, "!P"))
    {
	expires = 0;
	j = 1;
	perm = 1;

	if (strstr (orig, "!R"))
	    tmp = strtok (NULL, " ");
    }

    if (!stricmp (tmp, "!R"))
    {
	realname = 1;
	j = 1;

	if (strstr (orig, "!T"))
	    tmp = strtok (NULL, " ");
    }

    if (!stricmp (tmp, "!T"))
    {
	tmp = strtok (NULL, " ");
	expires = dotime (tmp);
	j = 1;
	if (expires <= 0)
	    expires = def_akill_time;
    }

    if (!j)
	memcpy (reason, orig, sizeof (reason));
    else
    {
	tmp = strtok (NULL, "");

	if (tmp)
	    memcpy (reason, strtok (NULL, ""), sizeof (reason));
	else
	    memcpy (reason, def_akill_reason, sizeof (reason));
    }

    /* Is the param a nick? If so, get the mask for it. But not if it's
       a realname akill...
     */
    if (!realname && (u = finduser (param)))
	snprintf (mask, sizeof (mask), "*@%s", u->host);
    else
	strscpy (mask, param, BUFSIZE);

    if (mask && reason)
    {
	if (is_akilled (mask))
	{
	    notice (s_RootServ, source, ERR_LIST_THERE, mask, "the", "AKill");
	    return;
	}

	/* Sanity checks - These don't apply if it's a realname akill */
	if ((strchr (mask, '!') && !realname) ||
	    (!strchr (mask, '@') && !realname) ||
	    (mask[strlen(mask)-1] == '@' && !realname))
	{
	    notice (s_RootServ, source, RS_AKILL_FORMAT, "");
	    return;
	}

	if (mask[0] == '@')
	{
	    char *mymask;
	    mymask = smalloc (strlen (mask) + 2);
	    strcpy (mymask, "*");
	    strcat (mymask, mask);
	    strcpy (mask, mymask);
	    free (mymask);
        }

	/* Make sure theres a sufficient number of non-wildcard characters.
	   @ and . are counted as 'wildcards' .. SRAs and realname akills
	   are exempt from this.
	 */
	if (!is_sra (source) && !realname)
	{
	    for (i = 0; i < strlen (mask); i++)
		if (!(mask[i] == '*' || mask[i] == '?' || mask[i] == '@' ||
		      mask[i] == '.'))
		    nonwild++;

	    if (nonwild < nonwildreq)
	    {
		notice (s_RootServ, source, ERR_TOOMANYWILD, nonwildreq);
		return;
	    }
	}

	/* Passed all the checks, now add it to the AKill list */
	akill = scalloc (sizeof (AutoKill), 1);
	akill->mask = sstrdup (mask);
	akill->reason = sstrdup (reason);
	akill->setter = sstrdup (source);
	akill->realname = realname;
	akill->set = time (NULL);
	akill->expires = expires;
	list = &akilllist[AHASH(akill->mask)];
	akill->next = *list;
	if (*list)
	    (*list)->prev = akill;
        *list = akill;

	notice (s_RootServ, source, RS_AKILL_ADDED, mask);
	if (globalonakill_on == TRUE)
	    globops (s_RootServ, RS_AKILL_ADDED_GLOBOP, source,
		realname ? "Realname " : "", perm ? "P" : "", mask, reason);

	if (!realname)
	{
	    username = sstrdup (akill->mask);
	    hostname = strchr (username, '@');

	    if (!hostname)
		return;

	    *hostname++ = 0;

#ifdef IRCUP9
	    /* We use 999999999 because we'll expire it ourselves. */
	    send_cmd (me.name, "GLINE * +%s@%s 999999999 :%s",
		username, hostname, akill->reason);
#elif defined (BAHAMUT) || defined (PROMETHEUS)
	    send_cmd (me.name, "AKILL %s %s 0 %s %ld :%s", hostname,
		username, s_RootServ, time (NULL), akill->reason);
#else
	    send_cmd (me.name, "%s %s %s :%s", me.token ? "V" : "AKILL",
		hostname, username, akill->reason);
#endif
	    free (username);
	}
	else
	{
	    int j = 0;

	    /* Hunt down any users that match this AKill. */
	    while (j < 1023)
	    {
		j++;
		u = userlist[j];
		if (u && match_wild_nocase (mask, u->real))
		{
#ifdef IRCUP9
		    /* We use 999999999 because we'll expire it ourselves. */
		    send_cmd (me.name, "GLINE * +*@%s 999999999 :%s",
			u->host, akill->reason);
#elif defined (BAHAMUT) || defined (PROMETHEUS)
		    send_cmd (me.name, "AKILL %s * 0 %s %ld :%s", u->host,
			s_RootServ, time (NULL), akill->reason);
#else
		    send_cmd (me.name, "%s %s * :%s", me.token ? "V" :
			"AKILL", u->host, akill->reason);
#endif
		    u = u->next;
		}
	    }
	}
    }
    else
    {
	notice (s_RootServ, source, RPL_SYNTAX,
	    "AKILL ADD Mask [Expiry] Reason");
	errmoreinfo (s_RootServ, source, "AKILL");
    }
}

/* Delete an AKill */
static void do_akill_del (const char *source)
{
    AutoKill *akill;
    int i, j = 0;
    char *param = strtok (NULL, " ");
#ifndef IRCUP9
    char *user, *host;
#endif

    if (!param)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "AKILL DEL Number|ALL");
	errmoreinfo (s_RootServ, source, "AKILL");
	return;
    }

    if (!stricmp (param, "ALL"))
    {
	for (i = 0; i < 1024; i++)
	{
	    for (akill = akilllist[i]; akill; akill = akill->next)
	    {
		j++;

		free (akill->setter);
		free (akill->reason);

#ifdef IRCUP9
		send_cmd (me.name, "GLINE * -%s", akill->mask);
#else
		user = sstrdup (akill->mask);
		host = strchr (user, '@');

		if (host)
		{
		    *host++ = 0;

		    send_cmd (me.name, "%s %s %s", me.token ? "Y" : "RAKILL",
			host, user);
		}

		free (user);
#endif /* IRCUP9 */

		if (akill->prev)
		    akill->prev->next = akill->next;
		else
		    akilllist[AHASH(akill->mask)] = akill->next;
		if (akill->next)
		    akill->next->prev = akill->prev;
	    }
	}

	free (akill);
	notice (s_RootServ, source, RS_ALL_AKILL_DELETED, j);

	if (globalonakill_on == TRUE)
	    globops (s_RootServ, RS_ALL_AKILL_DEL_GLOBOP, source);

	return;
    }
    else
    {
	for (i = 0; i < 1024; i++)
	{
	    for (akill = akilllist[i]; akill; akill = akill->next)
	    {
		j++;

		if (j == atoi (param) || !stricmp (param, akill->mask))
		{
		    notice (s_RootServ, source, RS_AKILL_DELETED, akill->mask);

		    if (globalonakill_on == TRUE)
			globops (s_RootServ, RS_AKILL_DELETED_GLOBOP, source,
			    akill->realname ? "Realname " : "",
			    akill->expires == 0 ? "P" : "", akill->mask);

		    free (akill->setter);
		    free (akill->reason);

#ifdef IRCUP9
		    send_cmd (me.name, "GLINE * -%s", akill->mask);
#else
		    user = sstrdup (akill->mask);
		    host = strchr (user, '@');

		    if (host) 
		    {
			*host++ = 0;

			send_cmd (me.name, "%s %s %s", me.token ? "Y" :
			    "RAKILL", host, user);
		    }

		    free (user);
#endif /* IRCUP9 */

		    if (akill->prev)
			akill->prev->next = akill->next;
		    else
			akilllist[AHASH(akill->mask)] = akill->next;
		    if (akill->next)
			akill->next->prev = akill->prev;

		    free (akill);

		    return;
		}
	    }
	}
    }
    notice (s_RootServ, source, RS_AKILL_NOT_FOUND, param);
}

/* List all AKills, or a specific akill */
static void do_akill_list (const char *source)
{
    AutoKill *akill;
    int i, j = 0, showreasons = 0;
    char *param1 = strtok (NULL, " ");
    char *param2 = strtok (NULL, " ");

    /* We only show reasons if asked to. */
    if (param1 && !stricmp (param1, "FULL"))
	showreasons = 1;

    notice (s_RootServ, source, RS_AKILL_LIST,
	showreasons ? " - With reasons" : "",
	(showreasons && param2) ? " - Matching " :
	(!showreasons && param1) ? " - Matching " : "",
	(showreasons && param2) ? param2 : (!showreasons && param1) ?
	    param1 : "");

    for (i = 0; i < 1024; i++)
    {
	for (akill = akilllist[i]; akill; akill = akill->next)
	{
	    if (showreasons && param2)
		if (!match_wild_nocase (param2, akill->mask) &&
		    !match_wild_nocase (param2, akill->setter))
		    continue;

	    if (!showreasons && param1)
		if (!match_wild_nocase (param1, akill->mask) &&
		    !match_wild_nocase (param1, akill->setter))
		    continue;

	    j++;

	    if (akill->expires == 0)
		notice (s_RootServ, source, RS_PAKILL_LIST_MATCH, j,
		    akill->mask, akill->setter,
		    showreasons ? akill->reason : "");
	    else
		notice (s_RootServ, source, RS_AKILL_LIST_MATCH, j,
		    akill->mask, akill->setter,
		    duration (akill->expires - (time (NULL) - akill->set), 1),
		    showreasons ? akill->reason : "");
	}
    }

    notice (s_RootServ, source, RS_AKILL_LIST_END);
}

/* Check all AKills and expire any nessecary */
void expire_akills ()
{
    AutoKill *akill;
#ifndef IRCUP9
    char *user, *host;
#endif
    int i, j = 0;

    if (logupdates_on == TRUE)
	log_sameline (1, "RS:SYNC: Checking expires... ");

    for (i = 0; i < 1024; i++)
    {
	for (akill = akilllist[i]; akill; akill = akill->next)
	{
	    /* AKills with an expire time of 0 are permanent */
	    if (akill->expires > 0)
	    {
		if ((akill->expires - (time (NULL) - akill->set)) < 0)
		{
		    log ("RS:SYNC:EXPIRE: %sAKill on %s by %s: %s",
			akill->realname ? "Realname " : "",
			akill->mask, akill->setter, akill->reason);

#ifdef IRCUP9
		    send_cmd (me.name, "GLINE * -%s", akill->mask);
#else
		    user = sstrdup (akill->mask);
		    host = strchr (user, '@');

		    if (host)
		    {
			*host++ = 0;

			send_cmd (me.name, "%s %s %s", me.token ? "Y" :
			    "RAKILL", host, user);
		    }

		    free (user);
#endif /* IRCUP9 */

		    free (akill->setter);
		    free (akill->reason);

		    if (akill->prev)
			akill->prev->next = akill->next;
		    else
			akilllist[AHASH(akill->mask)] = akill->next;
		    if (akill->next)
			akill->next->prev = akill->prev;

		    free (akill);

		    j++;
		}
	    }
	}
    }

    /* Only log this if we have something to report, or if we should
       log it regardless.
     */
    if (logupdates_on == TRUE)
	log_sameline (0, "%d akills expired.", j);
    else if (logupdates_on == FALSE && j > 0)
	log ("RS:SYNC:EXPIRE: %d AKills expired.", j);
}

/* If the given information matches an AKill, kill that user and place an
   AKill on it. (If we see a user that's AKilled, then the AKill was
   probably rehashed out or the server restarted)
 */
int check_akill (const char *nick, const char *user, const char *host,
		 const char *real)
{
    AutoKill *akill;
    int i;
    char buf[512];

    strscpy (buf, user, sizeof (buf) -2);
    i = strlen (buf);
    buf[i++] = '@';
    strscpy (buf + i, host, sizeof (buf) -i);

    for (i = 0; i < 1024; i++)
	for (akill = akilllist[i]; akill; akill = akill->next)
	{
	    if ((!akill->realname && match_wild_nocase (akill->mask, buf)) ||
		 (akill->realname && match_wild_nocase (akill->mask, real)))
	    {
		/* They're AKilled! For some reason the server hasn't disconnected
		   them yet. Perhaps our AKill was rehashed out. We'll re place it.
		   The server should disconnect them shortly after this. We'll send
		   out a kill, just in case....
		 */
#ifdef IRCUP9
		/* We use 999999999 because we'll expire it ourselves. */
		send_cmd (me.name, "GLINE * +%s@%s 999999999 :%s",
		    user, host, akill->reason);
#elif defined (BAHAMUT) || defined (PROMETHEUS)
		send_cmd (me.name, "AKILL %s %s 0 %s %ld :%s", host,
		    user, s_RootServ, time (NULL), akill->reason);
#else
		send_cmd (me.name, "%s %s %s :%s", me.token ? "V" : "AKILL",
		    host, user, akill->reason);
#endif

		kill_user (s_RootServ, nick, akill->reason);

		return 1;
	    }
	}

    return 0;
}

/* Display the list of clones on the network */
static void do_clones (const char *source)
{
    Clone *clone;
    User *user;
    char *param = strtok (NULL, " ");
    char *limit = strtok (NULL, " ");
    int i, j, max = 250, shown = 0, count = 0;

    if (!param)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "CLONES Pattern [Limit]");
	errmoreinfo (s_RootServ, source, "CLONES");
	return;
    }

    if (limit)
    {
	max = atoi (limit);

	if (max > 250)
	    max = 250;
    }

    notice (s_RootServ, source, RS_CLONES_START, max);
    notice (s_RootServ, source, RS_CLONES_HEADER);

    for (i = 0; i < 1024; i++)
    {
	for (clone = clonelist[i]; clone; clone = clone->next)
	{
	    int shost = 0;

	    if (!match_wild_nocase (param, clone->host))
		continue;

	    if (clone->cnt < 2)
		continue;

	    if (count >= max)
		break;

	    ++count;

	    for (j = 0; j < 1024; j++)
	    {
		for (user = userlist[j]; user; user = user->next)
		{
		    if (stricmp (clone->host, user->host))
			continue;

		    /* Very few hosts will go over this.. */
		    notice (s_RootServ, source, "%s%-44s %s%s%s%s%s",
			shost ? "" : "\2", shost ? "" : clone->host,
			user->nick, shost ? "" : " (",
			shost ? "" : itoa (clone->cnt), shost ? "" : ")",
			shost ? "" : "\2");
		    shost = 1;
		    shown++;
		}
	    }
	}
    }

    notice (s_RootServ, source, RS_CLONES_END, shown, count);
}

/* Handle a TRIGGER command */
static void do_trigger (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_RootServ, source, RPL_SYNTAX,
	    "TRIGGER ADD|DEL|LIST [Host|Number] [Limit]");
	errmoreinfo (s_RootServ, source, "TRIGGER");
    }

    else
    {
	Hash_SUB *command, hash_table[] =
	{
	    {"ADD",		H_IRCOP,	do_trigger_add},
	    {"DEL",		H_IRCOP,	do_trigger_del},
	    {"LIST",		H_IRCOP,	do_trigger_list},
	    {NULL}
	};

	if ((command = get_sub_hash (s_RootServ, source, strupper (cmd),
				 hash_table)))
	    (*command->process) (source);
	else
	{
	    notice (s_RootServ, source, RPL_SYNTAX,
		"TRIGGER ADD|DEL|LIST [Host|Number] [Limit]");
	    errmoreinfo (s_RootServ, source, "TRIGGER");
	}
    }
}

/* Add a trigger to the list */
static void do_trigger_add (const char *source)
{
    Trigger *trigger;
    char *mask = strtok (NULL, " "), *limit = strtok (NULL, " ");
    char *username, *hostname;

    if (!mask || !limit)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "TRIGGER ADD Host Limit");
	errmoreinfo (s_RootServ, source, "TRIGGER");
	return;
    }

    if (!strchr (mask, '@'))
    {
	notice (s_RootServ, source, RS_TRIGGER_FORMAT);
	return;
    }

    if (!is_sra (source))
    {
	int i, nonwild = 0;

	for (i = 0; i < strlen (mask); i++)
	    if (!(mask[i] == '*' || mask[i] == '?' || mask[i] == '@' ||
		mask[i] == '.'))
		nonwild++;

	if (nonwild < nonwildreq)
	{
	    notice (s_RootServ, source, ERR_TOOMANYWILD, nonwildreq);
	    return;
	}
    }

    username = sstrdup (mask);
    hostname = strchr (username, '@');

    if (!hostname)
	return;

    *hostname++ = 0;

    if ((trigger = findtrig (username, hostname)))
    {
	trigger->limit = atoi (limit);
	notice (s_RootServ, source, RS_RETRIGGER, mask, atoi (limit));
	free (username);
	return;
    }

    free (username);

    new_trigger (mask, atoi (limit));
    notice (s_RootServ, source, RS_TRIGGER_ADDED, mask);
}

/* Remove a trigger. */
static void do_trigger_del (const char *source)
{
    Trigger *trigger;
    char *param = strtok (NULL, " ");
    int i = 0;

    if (!param)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "TRIGGER DEL Number|Host|ALL");
	errmoreinfo (s_RootServ, source, "TRIGGER");
	return;
    }

    if (!stricmp (param, "ALL"))
    {
	for (trigger = triggerlist; trigger; trigger = trigger->next)
	{
	    i++;

	    /* Notice them up here so we can show the host */
	    notice (s_RootServ, source, RS_ALL_TRIGGER_DELETED, trigger->host);

	    if (trigger->prev)
		trigger->prev->next = trigger->next;
	    else
		triggerlist = trigger->next;
	    if (trigger->next)
		trigger->next->prev = trigger->prev;
	}

	free (trigger);
	return;
    }
    else
    {
	for (trigger = triggerlist; trigger; trigger = trigger->next)
	{
	    i++;
	    if (!stricmp (param, trigger->host) || i == atoi (param))
	    {
		if (trigger->prev)
		    trigger->prev->next = trigger->next;
		else
		    triggerlist = trigger->next;
		if (trigger->next)
		    trigger->next->prev = trigger->prev;

		free (trigger);
		if (atoi (param) > 0)
		    notice (s_RootServ, source, RS_TRIGGER_DELETED, "#", param);
		else
		    notice (s_RootServ, source, RS_TRIGGER_DELETED, "on",
			param);
		return;
	    }
	}
    }
    notice (s_RootServ, source, RS_TRIGGER_NOT_FOUND, param);
}

/* List the current triggers */
static void do_trigger_list (const char *source)
{
    Trigger *trigger;
    int i = 0;

    notice (s_RootServ, source, RS_TRIGGER_LIST_START);

    for (trigger = triggerlist; trigger; trigger = trigger->next)
    {
	i++;
	notice (s_RootServ, source, RS_TRIGGER_LIST, i, trigger->host,
	    trigger->limit);
    }

    notice (s_RootServ, source, RS_TRIGGER_LIST_END);
}

/* Handle a EXCEPTION command */
static void do_exception (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_RootServ, source, RPL_SYNTAX,
	    "EXCEPTION ADD|DEL|LIST [Host|Number]");
	errmoreinfo (s_RootServ, source, "EXCEPTION");
    }

    else
    {
	Hash_SUB *command, hash_table[] =
	{
	    {"ADD",		H_IRCOP,	do_exception_add},
	    {"DEL",		H_IRCOP,	do_exception_del},
	    {"LIST",		H_IRCOP,	do_exception_list},
	    {NULL}
	};

	if ((command = get_sub_hash (s_RootServ, source, strupper (cmd),
				 hash_table)))
	    (*command->process) (source);
	else
	{
	    notice (s_RootServ, source, RPL_SYNTAX,
		"EXCEPTION ADD|DEL|LIST [Host|Number]");
	    errmoreinfo (s_RootServ, source, "EXCEPTION");
	}
    }
}

/* Add a exception to the list */
static void do_exception_add (const char *source)
{
    Exception *exception;
    char *mask = strtok (NULL, " ");
    char *username, *hostname;

    if (!mask)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "EXCEPTION ADD Host");
	errmoreinfo (s_RootServ, source, "EXCEPTION");
	return;
    }

    if (!strchr (mask, '@'))
    {
	notice (s_RootServ, source, RS_EXCEPTION_FORMAT);
	return;
    }

    if (!is_sra (source))
    {
	int i, nonwild = 0;

	for (i = 0; i < strlen (mask); i++)
	    if (!(mask[i] == '*' || mask[i] == '?' || mask[i] == '@' ||
		mask[i] == '.'))
		nonwild++;

	if (nonwild < nonwildreq)
	{
	    notice (s_RootServ, source, ERR_TOOMANYWILD, nonwildreq);
	    return;
	}
    }

    username = sstrdup (mask);
    hostname = strchr (username, '@');

    if (!hostname)
	return;

    *hostname++ = 0;

    if ((exception = findexcept (username, hostname)))
    {
	notice (s_RootServ, source, RS_ALREADY_EXCEPT, mask);
	free (username);
	return;
    }

    free (username);

    new_exception (mask);
    notice (s_RootServ, source, RS_EXCEPTION_ADDED, mask);
}

/* Remove a exception. */
static void do_exception_del (const char *source)
{
    Exception *exception;
    char *param = strtok (NULL, " ");
    int i = 0;

    if (!param)
    {
	notice (s_RootServ, source, RPL_SYNTAX,
	    "EXCEPTION DEL Number|Host|ALL");
	errmoreinfo (s_RootServ, source, "EXCEPTION");
	return;
    }

    if (!stricmp (param, "ALL"))
    {
	for (exception = exceptionlist; exception; exception = exception->next)
	{
	    i++;

	    /* Notice them up here so we can show the host */
	    notice (s_RootServ, source, RS_EXCEPTIONS_DELETED,
		exception->host);

	    if (exception->prev)
		exception->prev->next = exception->next;
	    else
		exceptionlist = exception->next;
	    if (exception->next)
		exception->next->prev = exception->prev;
	}

	free (exception);
	return;
    }
    else
    {
	for (exception = exceptionlist; exception; exception = exception->next)
	{
	    i++;
	    if (!stricmp (param, exception->host) || i == atoi (param))
	    {
		if (exception->prev)
		    exception->prev->next = exception->next;
		else
		    exceptionlist = exception->next;
		if (exception->next)
		    exception->next->prev = exception->prev;

		free (exception);
		if (atoi (param) > 0)
		    notice (s_RootServ, source, RS_EXCEPTION_DELETED, "#",
			param);
		else
		    notice (s_RootServ, source, RS_EXCEPTION_DELETED, "on",
			param);
		return;
	    }
	}
    }
    notice (s_RootServ, source, RS_EXCEPTION_NOT_FOUND, param);
}

/* List the current exceptions */
static void do_exception_list (const char *source)
{
    Exception *exception;
    int i = 0;

    notice (s_RootServ, source, RS_EXCEPTION_LIST_START);

    for (exception = exceptionlist; exception; exception = exception->next)
    {
	i++;
	notice (s_RootServ, source, RS_EXCEPTION_LIST, i, exception->host);
    }

    notice (s_RootServ, source, RS_EXCEPTION_LIST_END);
}

/* Handle a CSOP command */
static void do_csop (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_RootServ, source, RPL_SYNTAX,
	    "CSOP ADD|DEL|LIST [NickName|Number|Pattern] [Max]");
	errmoreinfo (s_RootServ, source, "CSOP");
    }

    else
    {
	Hash_SUB *command, hash_table[] =
	{
	    {"ADD",		H_SRA,		do_csop_add},
	    {"DEL",		H_SRA,		do_csop_del},
	    {"LIST",		H_SRA,		do_csop_list},
	    {NULL}
	};

	if ((command = get_sub_hash (s_RootServ, source, strupper (cmd),
				     hash_table)))
	    (*command->process) (source);
	else
	{
	    notice (s_RootServ, source, RPL_SYNTAX,
		"CSOP ADD|DEL|LIST [NickName|Number|Pattern] [Max]");
	    errmoreinfo (s_RootServ, source, "CSOP");
	}
    }
}

/* Add a nick to the CSOp list */
static void do_csop_add (const char *source)
{
    NickInfo *ni;
    char *nick = strtok (NULL, " ");

    if (!nick)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "CSOP ADD NickName");
	errmoreinfo (s_RootServ, source, "CSOP");
	return;
    }

    ni = findnick (nick);

    if (!ni)
    {
	notice (s_RootServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    else if (ni->flags & NF_WAITAUTH)
    {
	notice (s_RootServ, source, NS_NOT_AUTHED, ni->nick);
	return;
    }

    else
    {
	if (ni->flags & NF_CSOP)
	{
	    notice (s_RootServ, source, ERR_LIST_THERE, ni->nick, "the", "CSOp");
	    return;
	}
	else
	{
	    ni->flags |= NF_CSOP;
	    notice (s_RootServ, source, RS_CSOP_ADDED, ni->nick);
	    return;
	}
    }
}

/* Remove a nick to the CSOp list */
static void do_csop_del (const char *source)
{
    NickInfo *ni;
    char *nick = strtok (NULL, " ");
    int i, j = 0;

    if (!nick)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "CSOP DEL NickName|Number");
	errmoreinfo (s_RootServ, source, "CSOP");
	return;
    }

    ni = findnick (nick);

    if (!ni)
    {
	/* It might be a number.. */
	for (i = 0; i < 1024; i++)
	{
	    for (ni = nicklist[i]; ni; ni = ni->next)
	    {
		if (ni->flags & NF_CSOP)
		{
		    j++;

		    if (j == atoi (nick))
		    {
			ni->flags &= ~NF_CSOP;
			notice (s_RootServ, source, RS_CSOP_REMOVED, ni->nick);
			return;
		    }
		}
	    }
	}

	notice (s_RootServ, source, ERR_LIST_NOT_THERE, nick, "the CSOp");
	return;
    }
    else
    {
	if (!(ni->flags & NF_CSOP))
	{
	    notice (s_RootServ, source, ERR_LIST_NOT_THERE, ni->nick, "the CSOp");
	    return;
	}
	else
	{
	    ni->flags &= ~NF_CSOP;
	    notice (s_RootServ, source, RS_CSOP_REMOVED, ni->nick);
	    return;
	}
    }
}

/* List the CSOps */
static void do_csop_list (const char *source)
{
    NickInfo *ni;
    char *pattern = strtok (NULL, " ");
    char *max = strtok (NULL, " ");
    int i, j = 0, nmax = 250;

    if (!pattern)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "CSOP LIST Pattern [Max]");
	errmoreinfo (s_RootServ, source, "CSOP");
	return;
    }

    if (max)
	nmax = atoi (max);
    if (nmax > 250)
	nmax = 250;

    notice (s_RootServ, source, RS_CSOP_LIST_START);

    for (i = 0; i < 1024; i++)
    {
	for (ni = nicklist[i]; ni; ni = ni->next)
	{
	    if (ni->flags & NF_CSOP && (match_wild_nocase (ni->nick, pattern) ||
		match_wild_nocase (ni->usermask, pattern)))
	    {
		j++;

		notice (s_RootServ, source, RS_CSOP_LIST, j, ni->nick, ni->usermask);

		if (j >= nmax)
		    break;
	    }
	}
    }

    notice (s_RootServ, source, RS_CSOP_LIST_END);
}

/* Load the RootServ DB from disk */
void load_rs_dbase (void)
{
    FILE *f = fopen (ROOTSERV_DB, "r");
    char *item, dBuf[2048], backup [2048];
    int log = 0, ain = 0, tin = 0, ein = 0;
#ifdef HAVE_GETTIMEOFDAY
    struct timeval start, now, tmp;
#endif

#ifdef HAVE_GETTIMEOFDAY
    log_sameline (1, "RS:DB: Loading... ");
#else
    log ("RS:DB: Loading...");
#endif

    if (!f)
    {
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (0, "No %s found!", ROOTSERV_DB);
#else
	log ("RS:DB: No %s found!", ROOTSERV_DB);
#endif

	log = 1;

	/* Check for a db.save */
	strcpy (backup, ROOTSERV_DB);
	strcat (backup, ".save");
	if (rename (backup, ROOTSERV_DB) > 0)
	{
#ifdef HAVE_GETTIMEOFDAY
	    log_sameline (0, "Recovering %s.save", ROOTSERV_DB);
#else
	    log ("RS:DB: Recovering %s.save", ROOTSERV_DB);
#endif
	    f = fopen (ROOTSERV_DB, "r");
	}
	else
	    return;
    }

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&(start), NULL);
#endif

    while (fgets (dBuf, 2047, f))
    {
	item = strtok (dBuf, " ");

	/* Ignore certain things.. */
	if (*item == '#' || *item == '\n' || *item == '\t' || *item == ' ' ||
	    *item == '\0' || *item == '\r')
	    continue;

	if (!strcmp (item, "AK"))
	{
	    AutoKill *akill, **list;

	    akill = scalloc (sizeof (AutoKill), 1);
	    akill->mask = sstrdup (strtok (NULL, " "));
	    akill->setter = sstrdup (strtok (NULL, " "));
	    akill->realname = atoi (strtok (NULL, " "));
	    akill->set = atol (strtok (NULL, " "));
	    akill->expires = atol (strtok (NULL, " "));
	    akill->reason = sstrdup (strtok (NULL, "\n"));
	    list = &akilllist[AHASH(akill->mask)];
	    akill->next = *list;
	    if (*list)
		(*list)->prev = akill;
	    *list = akill;

	    ain++;

	    continue;
	}

	if (!strcmp (item, "TR"))
	{
	    char *host;
	    int limit;

	    host = strtok (NULL, " ");
	    limit = atoi (strtok (NULL, "\n"));

	    new_trigger (host, limit);

	    tin++;

	    continue;
	}

	if (!strcmp (item, "EX"))
	{
	    char *host;

	    host = strtok (NULL, "\n");

	    new_exception (host);

	    ein++;

	    continue;
	}

	if (!strcmp (item, "NW"))
	{
	    news_time = atol (strtok (NULL, "\n"));
	    continue;
	}
    }

    fclose (f);

#ifdef HAVE_GETTIMEOFDAY
    if (!log)
    {
	gettimeofday (&now, NULL);
	timersub (&now, &start, &tmp);
	log_sameline (0, "finished in %d ms. (%d akills, %d triggers, %d exceptions)", tv2ms (&tmp), ain, tin, ein);
    }
#endif
}

/* Write the RootServ db to disk. */
void save_rs_dbase ()
{
    FILE *f;
    AutoKill *akill;
    Trigger *trigger;
    Exception *exception;
    int i, log = 0, aout = 0, tout = 0, eout = 0;
    char backup[2048];
#ifdef HAVE_GETTIMEOFDAY
    struct timeval start, now, tmp;
#endif

    if (logupdates_on == TRUE)
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (1, "RS:DB: Saving... ");
#else
	log ("RS:DB: Saving...");
#endif

    strcpy (backup, ROOTSERV_DB);
    strcat (backup, ".save");
    remove (backup);

    if (rename (ROOTSERV_DB, backup) < 0)
    {
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (0, "Can't back up %s!", ROOTSERV_DB);
#else
	log ("RS:DB: Can't back up %s!", ROOTSERV_DB);
#endif
	log = 1;
    }

    f = fopen (ROOTSERV_DB, "w");

    if (!f)
    {
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (0, "Can't write to %s!", ROOTSERV_DB);
#else
	log ("RS:DB: Can't write to %s!", ROOTSERV_DB);
#endif
	log = 1;

	if (rename (backup, ROOTSERV_DB) < 0)
	    fatal ("RS:DB: Can't restore backup of %s!", ROOTSERV_DB);
	return;
    }

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&(start), NULL);
#endif

    /* Write any AKills: One line per AKill. Format:
       AK mask setter realname? set expires reason
     */
    for (i = 0; i < 1024; i++)
	for (akill = akilllist[i]; akill; akill = akill->next)
	{
	    fprintf (f, "AK %s %s %d %lu %lu %s\n", akill->mask, akill->setter,
		akill->realname, akill->set, akill->expires, akill->reason);

	    aout++;
	}

    /* Write any Triggers: One line per trigger. Format:
       TR host limit
     */
    for (trigger = triggerlist; trigger; trigger = trigger->next)
    {
	fprintf (f, "TR %s %d\n", trigger->host, trigger->limit);

	tout++;
    }

    /* Write any Exceptions: One line per exception. Format:
       EX host
     */
    for (exception = exceptionlist; exception; exception = exception->next)
    {
	fprintf (f, "EX %s\n", exception->host);

	eout++;
    }

    /* Time of last NEWS Update */
    fprintf (f, "NW %lu\n", news_time);

    fclose (f);
    remove (backup);

#ifdef HAVE_GETTIMEOFDAY
    if (logupdates_on == TRUE && !log)
    {
	gettimeofday (&now, NULL);
	timersub (&now, &start, &tmp);
	log_sameline (0, "finished in %d ms (%d akills, %d triggers, %d exceptions)", tv2ms (&tmp), aout, tout, eout);
    }
#endif
}

/* Authenticate for SRA access. */
void do_auth (const char *source)
{
    char *nick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    User *u;
    SRA *sra;
         
    if (!nick || !pass)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "AUTH NickName Password");
	errmoreinfo (s_RootServ, source, "AUTH");
	return;
    }

    u = finduser (source);
    sra = findsra (nick);

    if (!u)
	return;

    if (!sra)
    {
	notice (s_RootServ, source, ERR_NOT_SRA);
	return;
    }

    /* Avoid excessive logging. */
    if (is_sra (source))
    {
	notice (s_RootServ, source, ERR_ALREADY_SRA);
	return;
    }

    if (stricmp (sra->nick, nick) || strcmp (sra->pass, pass))
    {
	/* Passwords don't match, do passfail stuff */
	++u->passfail;

	/* If they've exceeded the limit, kill them. */
	if (u->passfail == passfails)
	{
	    snoop (s_RootServ, "RS:AUTH:BP: %s!%s@%s (Killed)",
		source, u->user, u->host);
	    log ("RS:AUTH:BP: %s!%s@%s (Killed)", source, u->user, u->host);
	    kill_user (s_RootServ, source, RPL_KILLED_PASSFAIL);
	    return;
	}

	snoop (s_RootServ, "RS:AUTH:BP: %s!%s@%s (%d/%d)",
	    source, u->user, u->host, u->passfail, passfails);
	log ("RS:AUTH:BP: %s!%s@%s (%d/%d)",
	    source, u->user, u->host, u->passfail, passfails);

	notice (s_RootServ, source, ERR_NOT_SRA);

	globops (s_RootServ, RS_FAILED_AUTH, source, u->user, u->host);

	return;
    }
    else
    {
	u->flags |= UF_SRA;

	if (globalonauth_on == TRUE)
	    globops (s_RootServ, RS_NOW_SRA_GLOBOP, source, u->user, u->host);

	notice (s_RootServ, source, RS_NOW_SRA);
#ifdef UNREAL3
	send_cmd (s_RootServ, "%s %s :is a Services Root Admin", me.token ?
	    "BA" : "SWHOIS", source);
#endif
	return;
    }
}

/* De-Authenticate from SRA access. */
void do_deauth (const char *source)
{
    User *u;

    u = finduser (source);

    if (u->flags & UF_SRA)
    {
	u->flags &= ~UF_SRA;
	notice (s_RootServ, source, RS_NO_LONGER_SRA);
	return;
    }

    /* We shouldn't get here, but just in case. */
    notice (s_RootServ, source, ERR_NOT_SRA);
}

#ifdef RAWINJECT
/* Send raw info to the uplink */
void do_raw (const char *source)
{
    char *text = strtok (NULL, "");

    if (!text)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "RAW Command");
	errmoreinfo (s_RootServ, source, "RAW");
    }
    else
	send_cmd (NULL, text);
}

/* Throw something into the inbuf and process it. */
void do_inject (const char *source)
{
    char *text = strtok (NULL, "");

    if (!text)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "INJECT Data");
	errmoreinfo (s_RootServ, source, "INJECT");
    }
    else
    {
	strscpy (inbuf, text, BUFSIZE);
	process();
    }
}
#endif /* RAWINJECT */

/* Toggle debug mode. */
void do_debug (const char *source)
{
    char *param = strtok (NULL, " ");
        
    if (!param)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "DEBUG Level");
	errmoreinfo (s_RootServ, source, "DEBUG");
	return;
    }

    if (atoi (param) > 0)
    {
	debuglevel = atoi (param);
	open_debug();
	debug ("-----------------------------------------------------");
	notice (s_RootServ, source, RS_DEBUG_LEVEL, atoi (param),
	    debuglevel == 1 ? "Server<->Server traffic" :
	    debuglevel == 2 ? "Some function calls" :
	    debuglevel == 3 ? "Server<->Server traffic and some function calls"
	    : "Unknown");
    }
    else
    {
	notice (s_RootServ, source, RS_DEBUG_LEVEL, atoi (param), "Off");
	debug ("%s turned debug mode off.", source);
	debuglevel = 0;
	close_debug();
    }
}

/* Terminate Services permanently. */
void do_shutdown (const char *source)
{
    char reason[BUFSIZE], orig[BUFSIZE], *param;

    param = strtok (NULL, "");

    if (!param)
	memcpy (reason, "No reason given", sizeof (reason));
    else
	memcpy (reason, param, sizeof (reason));

    strscpy (orig, reason, sizeof (orig));

    param = strtok (reason, " ");

    /* Should we skip the DB save? */
    if (!stricmp (param, "NOSAVE"))
    {
	/* Yep. */
	param = strtok (NULL, "");

	if (param)
	    snprintf (quitmsg, BUFSIZE, param);
	else
	    snprintf (quitmsg, BUFSIZE, "No reason given");

	globops (s_RootServ, RS_SHUTTING_DOWN, source, quitmsg);
	runflags |= RUN_SHUTDOWN;
    }
    else
    {
	/* Nope. Save the DBs. */
	snprintf (quitmsg, BUFSIZE, orig);

	expire_akills ();
	save_rs_dbase ();

	if (nickserv_on == TRUE)
	{
	    expire_nicks ();
	    save_ns_dbase ();
	}

	if (chanserv_on == TRUE)
	{
	    expire_chans ();
	    save_cs_dbase ();
	}

	globops (s_RootServ, RS_SHUTTING_DOWN, source, quitmsg);
	runflags |= RUN_SHUTDOWN;
    }
}

/* Restart Services. */
void do_restart (const char *source)
{
    char reason[BUFSIZE], orig[BUFSIZE], *param;

    param = strtok (NULL, "");

    if (!param)
	memcpy (reason, "No reason given", sizeof (reason));
    else
	memcpy (reason, param, sizeof (reason));

    strscpy (orig, reason, sizeof (orig));

    param = strtok (reason, " ");

    /* If wait_restart isn't set, Services won't restart. But if we're
       using this function, we want them to restart, so we set
       wait_restart to 5 seconds here. This change is wiped out when
       they restart.
     */
    wait_restart = 5;

    /* Should we skip the DB save? */
    if (!stricmp (param, "NOSAVE"))
    {
	/* Yep. */
	param = strtok (NULL, "");
	snprintf (quitmsg, BUFSIZE, param);

	if (!quitmsg)
	    snprintf (quitmsg, BUFSIZE, "No reason given");

	globops (s_RootServ, RS_RESTARTING, duration (wait_restart, 1), source,
	    quitmsg);
	runflags |= RUN_RESTART;
    }
    else
    {
	/* Nope. Save the DBs. */
	snprintf (quitmsg, BUFSIZE, orig);

	expire_akills ();
	save_rs_dbase ();

	if (nickserv_on == TRUE)
	{
	    expire_nicks ();
	    save_ns_dbase ();
	}

	globops (s_RootServ, RS_RESTARTING, duration (wait_restart, 1), source,
	    quitmsg);
	runflags |= RUN_RESTART;
    }
}

/* Show various Services settings */
void do_settings (const char *source)
{
    notice (s_RootServ, source, RS_SETTINGS_START);
    notice (s_RootServ, source, " ");
#ifdef UNREAL3
    notice (s_RootServ, source, RS_SETTINGS_UNREAL, unreal_numeric);
#endif
    notice (s_RootServ, source, RS_SETTINGS_NETWORK, network_name);

    if (sendmail_on == TRUE)
	notice (s_RootServ, source, RS_SETTINGS_SENDMAIL, sendmail_path);

    notice (s_RootServ, source, RS_SETTINGS_SYNC, duration (update_timeout, 1));
    notice (s_RootServ, source, RS_SETTINGS_TEMPEXPIRE, duration (tempexpire, 1));
    notice (s_RootServ, source, RS_SETTINGS_PASSFAILS, passfails);
    notice (s_RootServ, source, RS_SETTINGS_HAVESERV, haveserv_on == TRUE ? "On" : "Off");
    notice (s_RootServ, source, RS_SETTINGS_HELPINDEX, helpindex == 1 ? "Short" :
	helpindex == 2 ? "Medium" : helpindex == 3 ? "Long" : "Unknown");

    if (floodmsgs)
    {
	notice (s_RootServ, source, RS_SETTINGS_FLOOD, floodmsgs,
	    duration (floodtime, 1));
	notice (s_RootServ, source, RS_SETTINGS_NOISYFLOOD, noisyflood_on == TRUE ?
	    "On" : "Off");
    }

    if (strlen (snoopchan))
	notice (s_RootServ, source, RS_SETTINGS_SNOOPCHAN, snoopchan);

    if (strlen (joinchan))
	notice (s_RootServ, source, RS_SETTINGS_JOINCHAN, joinchan);

    if (globalnoticer_on == TRUE)
	notice (s_RootServ, source, RS_SETTINGS_TLDS, tlds);

    if (nickserv_on == TRUE)
    {
	notice (s_RootServ, source, RS_SETTINGS_NREGTYPE,
	    nsregistertype == 0 ? "Normal" :
	    nsregistertype == 1 ? "EMailAuth" :
	    nsregistertype == 2 ? "EMailNoAuth" :
	    nsregistertype == 3 ? "PassAuth" :
	    nsregistertype == 4 ? "Verify" :
	    nsregistertype == 5 ? "EMailVerify" :
	    "AuthVerify");
	notice (s_RootServ, source, RS_SETTINGS_SSTAMP, sstamp);
	notice (s_RootServ, source, RS_SETTINGS_NICKDELAY, duration (nick_delay, 1));
	notice (s_RootServ, source, RS_SETTINGS_MAXLINKS, maxlinks);
	notice (s_RootServ, source, RS_SETTINGS_EMAILLIMIT, emaillimit);
	notice (s_RootServ, source, RS_SETTINGS_NICKEXPIRE,
	    duration (nick_expire, 1));
	notice (s_RootServ, source, RS_SETTINGS_COLLIDE,
	    collidetype == 1 ? "Guest" : collidetype == 2 ? "User" :
	    collidetype == 3 ? "Nick" : collidetype == 4 ? "Underscore" :
	    collidetype == 5 ? "Hyphen" : "Kill");

	if (strlen (nsregister_url))
	    notice (s_RootServ, source, RS_SETTINGS_NSRURL,
		nsregister_url);

	if (strlen (collision_url))
	    notice (s_RootServ, source, RS_SETTINGS_NSCURL,
		collision_url);
    }

    if (memoserv_on == TRUE)
    {
	notice (s_RootServ, source, RS_SETTINGS_MEMOMAILMAX, memomailmax);
	notice (s_RootServ, source, RS_SETTINGS_MEMOUSERMAX, memosfromuser);
	notice (s_RootServ, source, RS_SETTINGS_MEMODELAY, duration (memo_delay, 1));

	if (strlen (memomail_url))
	    notice (s_RootServ, source, RS_SETTINGS_MURL,
		memomail_url);
    }

    if (chanserv_on == TRUE)
    {
	notice (s_RootServ, source, RS_SETTINGS_CREGTYPE,
	    csregistertype == 0 ? "Normal" :
	    csregistertype == 1 ? "EMailAuth" :
	    csregistertype == 2 ? "Verify" :
	    csregistertype == 3 ? "AuthVerify" :
	    "CSOpVerify");
	notice (s_RootServ, source, RS_SETTINGS_CHANLIMIT, chanlimit);
	notice (s_RootServ, source, RS_SETTINGS_CHANDELAY, duration (chan_delay, 1));
	notice (s_RootServ, source, RS_SETTINGS_CHANEXPIRE,
	    duration (chan_expire, 1));

	if (strlen (csregister_url))
	    notice (s_RootServ, source, RS_SETTINGS_CSRURL,
		csregister_url);
    }

    if (rootserv_on == TRUE)
    {
	notice (s_RootServ, source, RS_SETTINGS_SECURESRA,
	   securesra_on == TRUE ? "On" : "Off");
	notice (s_RootServ, source, RS_SETTINGS_AKILL, duration
	    (def_akill_time, 1));
	notice (s_RootServ, source, RS_SETTINGS_AKILLREASON, def_akill_reason);
	notice (s_RootServ, source, RS_SETTINGS_AKILLGLOBAL,
	    globalonakill_on == TRUE ? "On" : "Off");
	notice (s_RootServ, source, RS_SETTINGS_AKILLLEARN,
	    learnakills_on == TRUE ? "On" : "Off");
	notice (s_RootServ, source, RS_SETTINGS_AKILLFORGET,
	    forgetakills_on == TRUE ? "On" : "Off");
	notice (s_RootServ, source, RS_SETTINGS_NONWILD, nonwildreq);
	notice (s_RootServ, source, RS_SETTINGS_AUTHGLOBAL,
	    globalonauth_on == TRUE ? "On" : "Off");
    }

    if (globalnoticer_on == TRUE)
	if (strlen (globalmsg))
	    notice (s_RootServ, source, RS_SETTINGS_GLOBALMSG, globalmsg);

    notice (s_RootServ, source, RS_SETTINGS_LOGUPDATES,
	logupdates_on == TRUE ? "On" : "Off");
    notice (s_RootServ, source, RS_SETTINGS_STATGLOBAL,
	statglobal_on == TRUE ? "On" : "Off");
    notice (s_RootServ, source, RS_SETTINGS_CLONES, maxclones, maxclones ? "" : " (Off)");
    notice (s_RootServ, source, RS_SETTINGS_CLONEKILLS, maxclonekills,
	maxclonekills ? "" : " (Off)");
    notice (s_RootServ, source, RS_SETTINGS_RESTART,
	duration (wait_restart, 1));
    notice (s_RootServ, source, RS_SETTINGS_DEBUG, debuglevel,
	debuglevel ? "" : " (Off)");
    notice (s_RootServ, source, " ");
    notice (s_RootServ, source, RS_SETTINGS_END);
}

/* Update the DBs. */
void do_update (const char *source)
{
    globops (s_RootServ, RS_FORCE_UPDATE, source);
    runflags |= RUN_SAVEDBS;
}

/* Re-read services.conf */
void do_rehash (const char *source)
{
    /* Just in case. */
    do_update (source);

    globops (s_RootServ, RS_DOING_REHASH, CONFIGFILE, source);

    /* Remove all clients from the JoinChan. We do this here
       before we turn the service bots off. */
    if (strlen (joinchan))
	joinchan_part ();

    /* Turn these off. They'll be re-enabled if they're
       turned on in the conf.
     */
    rootserv_on = FALSE;
    nickserv_on = FALSE;
    chanserv_on = FALSE;
    memoserv_on = FALSE;
    globalnoticer_on = FALSE;
    globalonakill_on = FALSE;
    logupdates_on = FALSE;
    learnakills_on = FALSE;
    forgetakills_on = FALSE;
    age_on = FALSE;
    sex_on = FALSE;
    location_on = FALSE;
    name_on = FALSE;
    privmsg_on = FALSE;
    showmail_on = FALSE;
    *joinchan = 0;
    maxlinks = 0;

    /* Flush SRAs (some might be removed) */
    flushlists ();

    /* Now re-read. */
    load_conf ();
    
    /* Now check it */
    check_conf(); 

    /* Remove any disabled pseudoclients */
    if (rs.isonline && rootserv_on == FALSE)
    {
	send_cmd (s_RootServ, "%s :Client disabled", me.token ? "," : "QUIT");
	--userstat;
	rs.isonline = 0;
    }

    if (ns.isonline && nickserv_on == FALSE)
    {
	send_cmd (s_NickServ, "%s :Client disabled", me.token ? "," : "QUIT");
	--userstat;
	ns.isonline = 0;
    }

    if (cs.isonline && chanserv_on == FALSE)
    {
	send_cmd (s_ChanServ, "%s :Client disabled", me.token ? "," : "QUIT");
	--userstat;
	cs.isonline = 0;
    }

    if (ms.isonline && memoserv_on == FALSE)
    {
	send_cmd (s_MemoServ, "%s :Client disabled", me.token ? "," : "QUIT");
	--userstat;
	ms.isonline = 0;
    }

    if (gn.isonline && globalnoticer_on == FALSE)
    {
	send_cmd (s_GlobalNoticer, "%s :Client disabled", me.token ? "," :
	    "QUIT");
	--userstat;
	gn.isonline = 0;
    }

    /* Put all clients back in the JoinChan */
    if (strlen (joinchan))
	joinchan_join ();
}

/* Every time I think I've got the channels/user counts to stop
   getting desync'd from the network, I take these out .. and every
   time I do, some other new method of getting desync'd pops up. Feh.
 */
static void do_userlist (const char *source)
{
    User *u;
    struct u_chanlist *c;
    int i, j = 0;
    char buf[BUFSIZE];
 
    for (i = 0; i < 1024; i++)
    {
        for (u = userlist[i]; u; u = u->next)
        {
            j++;
            notice (s_RootServ, source, "%s (%s@%s)", u->nick,
                u->user, u->host);
                
            c = u->chans;
             
            *buf = 0;
                    
            for (c = u->chans; c; c = c->next)
            {
                strcat (buf, c->chan->name);
                strcat (buf, " ");
            }
 
            if (*buf)
                notice (s_RootServ, source, "%s", buf);
            else
                notice (s_RootServ, source, "None");
        }
    }
    notice (s_RootServ, source, "%d users", j);
}
            
static void do_chanlist (const char *source)
{
    Channel *c;
    char buf[BUFSIZE];
    int i, j = 0, k;
    struct c_userlist *u;
    
    for (i = 0; i < 1024; i++)
    {       
        for (c = chanlist[i]; c; c = c->next)
        {
            j++;
             
            *buf = 0;
                
            k = 0;
   
            for (u = c->users; u && u->user; u = u->next)
            {
                if (*buf)
                    strcat (buf, " ");
                strcat (buf, u->user->nick);
     
                k++;
            }
    
            notice (s_RootServ, source, "%s %s%s%s", c->name,
                c->mode & CMODE_s ? "(" : "",
                c->mode & CMODE_s ? "s" : "",
                c->mode & CMODE_s ? ")" : "");
    
            notice (s_RootServ, source, "%d %s", k, buf);
        }       
    }
             
    notice (s_RootServ, source, "%d channels", j);
}

/* Search the log file for the specified text. Return Num matching lines. */
void do_search (const char *source)
{
    FILE *log = fopen (LOGFILE, "r");
    char *match = strtok (NULL, " ");
    char *param = strtok (NULL, " ");
    char tmp[BUFSIZE];
    char dBuf[2048];
    int max, found = 0;

    if (!match)
    {
	notice (s_RootServ, source, RPL_SYNTAX, "SEARCH MatchText [Num]");
	errmoreinfo (s_RootServ, source, "SEARCH");
	return;
    }

    if (!param || (param && atoi (param) > 250))
	max = 50;
    else
	max = atoi (param);

    if (!log)
    {
	notice (s_RootServ, source, ERR_CANT_OPEN, LOGFILE);
	return; 
    }

    notice (s_RootServ, source, RS_LOG_START, max);

    snprintf (tmp, sizeof (tmp), "*%s*", match);

    while (fgets (dBuf, 2047, log))
    {
	if (found >= max)
	    break;

	if (match_wild_nocase (tmp, dBuf))
	{
	    ++found;
	    notice (s_RootServ, source, dBuf);
	}
    }

    notice (s_RootServ, source, RS_LOG_END, found);

    fclose (log);
}

/* Show various services statistics */
static void do_stats (const char *source)
{
    NickInfo *ni;
    MemoInfo *mi;
    ChanInfo *ci;
    AutoKill *akill;
    int i, nicks = 0, links = 0, chans = 0, memos = 0, akills = 0;
    int uptime = time (NULL) - me.since;

    notice (s_RootServ, source, RS_STATS_START);
    notice (s_RootServ, source, " ");
    notice (s_RootServ, source, RS_STATS_NETWORK);
    notice (s_RootServ, source, " ");
    notice (s_RootServ, source, RS_STATS_USERS, userstat);
    notice (s_RootServ, source, RS_STATS_CHANNELS, chancnt);
    notice (s_RootServ, source, RS_STATS_SERVERS, servcnt + 1);
    notice (s_RootServ, source, " ");
    notice (s_RootServ, source, RS_STATS_SERVICES);
    notice (s_RootServ, source, " ");

    for (ni = firstni (); ni; ni = nextni ())
    {
	if (ni->flags & NF_LINKED)
	    links++;
	else
	    nicks++;

	mi = &ni->memos;

	if (mi->memocnt)
	    memos += mi->memocnt;
    }

    notice (s_RootServ, source, RS_STATS_NICKINFO, nicks);
    notice (s_RootServ, source, RS_STATS_LINKED, links);
    notice (s_RootServ, source, RS_STATS_MEMOS,	memos);

    for (ci = firstci (); ci; ci = nextci ())
	chans++;

    notice (s_RootServ, source, RS_STATS_CHANINFO, chans);

    for (i = 0; i < 1024; i++)
	for (akill = akilllist[i]; akill; akill = akill->next)
	    akills++;

    notice (s_RootServ, source, RS_STATS_AKILLS, akills);

    notice (s_RootServ, source, RS_STATS_UPTIME, uptime/86400, (uptime/3600) % 24,
	(uptime/60) % 60, uptime % 60);

    notice (s_RootServ, source, " ");
    notice (s_RootServ, source, RS_STATS_END);
}
